@@ -57,10 +57,12 @@ struct _GisAssistantPrivate
   GtkWidget *frame;
   GtkWidget *forward;
   GtkWidget *back;
+  GtkWidget *cancel;
   GtkWidget *progress_indicator;
   GtkWidget *main_layout;
   GtkWidget *action_area;
   GtkWidget *page_action_widget_area;
+  GtkWidget *spinner;
 
   GList *pages;
   GisPage *current_page;
@@ -94,12 +96,28 @@ gis_assistant_switch_to (GisAssistant          *assistant,
   GIS_ASSISTANT_GET_CLASS (assistant)->switch_to (assistant, direction, page);
 }
 
+static void
+on_apply_done (GisPage *page,
+               gboolean valid,
+               gpointer user_data)
+{
+  GisAssistant *assistant = GIS_ASSISTANT (user_data);
+  if (valid)
+    g_signal_emit (assistant, signals[NEXT_PAGE], 0,
+                   assistant->priv->current_page);
+  g_object_unref (assistant);
+}
+
 void
 gis_assistant_next_page (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = assistant->priv;
-  g_signal_emit (assistant, signals[NEXT_PAGE], 0,
-                 priv->current_page);
+  if (priv->current_page)
+    gis_page_apply_begin (priv->current_page, on_apply_done,
+                          g_object_ref (assistant));
+  else
+    g_signal_emit (assistant, signals[NEXT_PAGE], 0,
+                   priv->current_page);
 }
 
 static inline gboolean
@@ -234,6 +252,24 @@ update_navigation_buttons (GisAssistant *assistant)
     }
 }
 
+static void
+update_applying_state (GisAssistant *assistant)
+{
+  gboolean applying = FALSE;
+  if (assistant->priv->current_page)
+    applying = gis_page_get_applying (assistant->priv->current_page);
+  gtk_widget_set_sensitive (assistant->priv->frame, !applying);
+  gtk_widget_set_sensitive (assistant->priv->forward, !applying);
+  gtk_widget_set_visible (assistant->priv->back, !applying);
+  gtk_widget_set_visible (assistant->priv->cancel, applying);
+  gtk_widget_set_visible (assistant->priv->spinner, applying);
+
+  if (applying)
+    gtk_spinner_start (GTK_SPINNER (assistant->priv->spinner));
+  else
+    gtk_spinner_stop (GTK_SPINNER (assistant->priv->spinner));
+}
+
 static void
 gis_assistant_real_prepare (GisAssistant *assistant,
                             GisPage      *page)
@@ -253,6 +289,8 @@ page_notify (GisPage      *page,
 
   if (strcmp (pspec->name, "title") == 0)
     g_object_notify_by_pspec (G_OBJECT (assistant), obj_props[PROP_TITLE]);
+  else if (strcmp (pspec->name, "applying") == 0)
+    update_applying_state (assistant);
   else
     update_navigation_buttons (assistant);
 }
@@ -309,6 +347,14 @@ go_backward (GtkWidget    *button,
   gis_assistant_previous_page (assistant);
 }
 
+static void
+do_cancel (GtkWidget    *button,
+           GisAssistant *assistant)
+{
+  if (assistant->priv->current_page)
+    gis_page_apply_cancel (assistant->priv->current_page);
+}
+
 gchar *
 gis_assistant_get_title (GisAssistant *assistant)
 {
@@ -333,6 +379,7 @@ _gis_assistant_current_page_changed (GisAssistant *assistant,
 
   if (priv->current_page != page) {
     priv->current_page = page;
+    update_applying_state (assistant);
     g_object_notify_by_pspec (G_OBJECT (assistant), obj_props[PROP_TITLE]);
     g_signal_emit (assistant, signals[PREPARE], 0, page);
   }
@@ -346,6 +393,7 @@ gis_assistant_locale_changed (GisAssistant *assistant)
 
   gtk_button_set_label (GTK_BUTTON (priv->forward), _("_Next"));
   gtk_button_set_label (GTK_BUTTON (priv->back), _("_Back"));
+  gtk_button_set_label (GTK_BUTTON (priv->cancel), _("_Cancel"));
 
   for (l = priv->pages; l != NULL; l = l->next)
     gis_page_locale_changed (l->data);
@@ -356,6 +404,8 @@ gis_assistant_init (GisAssistant *assistant)
 {
   GisAssistantPrivate *priv = GET_PRIVATE (assistant);
   GtkWidget *navigation;
+  GtkWidget *widget;
+
   assistant->priv = priv;
 
   priv->main_layout = gtk_box_new (GTK_ORIENTATION_VERTICAL, 20);
@@ -383,11 +433,25 @@ gis_assistant_init (GisAssistant *assistant)
                         gtk_image_new_from_stock (GTK_STOCK_GO_BACK, GTK_ICON_SIZE_BUTTON));
   gtk_button_set_use_underline (GTK_BUTTON (priv->back), TRUE);
 
+  priv->cancel = gtk_button_new ();
+  gtk_button_set_image (GTK_BUTTON (priv->cancel),
+                        gtk_image_new_from_stock (GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON));
+  gtk_button_set_use_underline (GTK_BUTTON (priv->cancel), TRUE);
+
+  priv->spinner = gtk_spinner_new ();
+  widget = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
+  gtk_alignment_set_padding (GTK_ALIGNMENT (widget), 0, 0, 12, 6);
+  gtk_box_pack_start (GTK_BOX (navigation), widget, FALSE, FALSE, 0);
+  gtk_container_add (GTK_CONTAINER (widget), priv->spinner);
+  gtk_widget_show (widget);
+
+  gtk_box_pack_start (GTK_BOX (navigation), priv->cancel, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (navigation), priv->back, FALSE, FALSE, 0);
   gtk_box_pack_start (GTK_BOX (navigation), priv->forward, FALSE, FALSE, 0);
 
   g_signal_connect (priv->forward, "clicked", G_CALLBACK (go_forward), assistant);
   g_signal_connect (priv->back, "clicked", G_CALLBACK (go_backward), assistant);
+  g_signal_connect (priv->cancel, "clicked", G_CALLBACK (do_cancel), assistant);
 
   priv->progress_indicator = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
   gtk_widget_set_halign (priv->progress_indicator, GTK_ALIGN_CENTER);
@@ -399,6 +463,7 @@ gis_assistant_init (GisAssistant *assistant)
   gis_assistant_locale_changed (assistant);
   gtk_box_pack_start (GTK_BOX (priv->main_layout), priv->action_area, FALSE, TRUE, 0);
 
+  update_applying_state (assistant);
   gtk_widget_show_all (GTK_WIDGET (assistant));
 }
 
