@@ -42,6 +42,17 @@
 #include <cheese-gtk.h>
 #endif
 
+static void        join_show_prompt    (GisAccountPage *page,
+                                        GError *error);
+
+static void        on_join_login       (GObject *source,
+                                        GAsyncResult *result,
+                                        gpointer user_data);
+
+static void        on_realm_joined     (GObject *source,
+                                        GAsyncResult *result,
+                                        gpointer user_data);
+
 G_DEFINE_TYPE (GisAccountPage, gis_account_page, GIS_TYPE_PAGE);
 
 #define GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GIS_TYPE_ACCOUNT_PAGE, GisAccountPagePrivate))
@@ -73,6 +84,11 @@ struct _GisAccountPagePrivate
   UmRealmManager *realm_manager;
   gboolean domain_chosen;
 
+  /* Valid during apply */
+  UmRealmObject *realm;
+  GCancellable *cancellable;
+  gboolean join_prompted;
+
   GtkWidget *action;
 };
 
@@ -86,6 +102,9 @@ show_error_dialog (GisAccountPage *page,
 {
   GtkWidget *dialog;
 
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    return;
+
   dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (page))),
                                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                                    GTK_MESSAGE_ERROR,
@@ -518,7 +537,7 @@ local_create_user (GisAccountPage *page)
   act_user_set_user_name (priv->act_user, username);
   act_user_set_account_type (priv->act_user, priv->account_type);
   if (strlen (password) == 0)
-    act_user_set_account_type (priv->act_user, ACT_USER_PASSWORD_MODE_NONE);
+    act_user_set_password_mode (priv->act_user, ACT_USER_PASSWORD_MODE_NONE);
   else
     act_user_set_password (priv->act_user, password, "");
 
@@ -556,12 +575,14 @@ on_permit_user_login (GObject *source,
     g_debug ("Caching remote user: %s", login);
 
     priv->act_user = act_user_manager_cache_user (priv->act_client, login, NULL);
+    gis_page_apply_complete (GIS_PAGE (page), TRUE);
 
     g_free (login);
   } else {
     show_error_dialog (page, _("Failed to register account"), error);
     g_message ("Couldn't permit logins on account: %s", error->message);
     g_error_free (error);
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
   }
 }
 
@@ -588,7 +609,7 @@ enterprise_permit_user_login (GisAccountPage *page, UmRealmObject *realm)
 
   um_realm_common_call_change_login_policy (common, "",
                                             add, remove, options,
-                                            NULL,
+                                            page->priv->cancellable,
                                             on_permit_user_login,
                                             page);
 
@@ -596,34 +617,213 @@ enterprise_permit_user_login (GisAccountPage *page, UmRealmObject *realm)
   g_free (login);
 }
 
+static void
+on_set_static_hostname (GObject *source,
+                        GAsyncResult *result,
+                        gpointer user_data)
+{
+  GisAccountPage *page = user_data;
+  GError *error = NULL;
+  GVariant *retval;
+  const gchar *name;
+
+  retval = g_dbus_connection_call_finish (G_DBUS_CONNECTION (source), result, &error);
+  if (error != NULL) {
+    join_show_prompt (page, error);
+    g_error_free (error);
+    return;
+  }
+
+  g_variant_unref (retval);
+
+  name = gtk_entry_get_text (OBJ (GtkEntry *, "join-name"));
+  g_debug ("Logging in as admin user: %s", name);
+
+  /* Prompted for some admin credentials, try to use them to log in */
+  um_realm_login (page->priv->realm, name,
+                  gtk_entry_get_text (OBJ (GtkEntry *, "join-password")),
+                  page->priv->cancellable, on_join_login, page);
+}
+
+static void
+on_join_response (GtkDialog *dialog,
+                  gint response,
+                  gpointer user_data)
+{
+  GisAccountPage *page = user_data;
+  GDBusConnection *connection;
+  GError *error = NULL;
+  gchar hostname[128];
+  const gchar *name;
+
+  gtk_widget_hide (GTK_WIDGET (dialog));
+  if (response != GTK_RESPONSE_OK) {
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
+    return;
+  }
+
+  name = gtk_entry_get_text (OBJ (GtkEntry *, "join-computer"));
+  if (gethostname (hostname, sizeof (hostname)) == 0 &&
+      !g_str_equal (name, hostname)) {
+    g_debug ("Setting StaticHostname to '%s'", name);
+
+    connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, page->priv->cancellable, &error);
+    if (error != NULL) {
+      gis_page_apply_complete (GIS_PAGE (page), FALSE);
+      g_warning ("Could not get DBus connection: %s", error->message);
+      g_error_free (error);
+      return;
+    }
+
+    g_dbus_connection_call (connection, "org.freedesktop.hostname1",
+                            "/org/freedesktop/hostname1", "org.freedesktop.hostname1",
+                            "SetStaticHostname",
+                            g_variant_new ("(sb)", name, TRUE),
+                            G_VARIANT_TYPE ("()"),
+                            G_DBUS_CALL_FLAGS_NONE,
+                            G_MAXINT, NULL, on_set_static_hostname, page);
+
+  } else {
+    name = gtk_entry_get_text (OBJ (GtkEntry *, "join-name"));
+    g_debug ("Logging in as admin user: %s", name);
+
+    /* Prompted for some admin credentials, try to use them to log in */
+    um_realm_login (page->priv->realm, name,
+                    gtk_entry_get_text (OBJ (GtkEntry *, "join-password")),
+                    NULL, on_join_login, page);
+  }
+}
+
+static void
+join_show_prompt (GisAccountPage *page,
+                  GError *error)
+{
+  UmRealmKerberosMembership *membership;
+  UmRealmKerberos *kerberos;
+  gchar hostname[128];
+  const gchar *name;
+  GtkDialog *join_dialog;
+  GtkEntry *join_name;
+  GtkEntry *join_password;
+  GtkLabel *join_domain;
+  GtkEntry *join_computer;
+
+  join_dialog = OBJ (GtkDialog *, "join-dialog");
+  join_domain = OBJ (GtkLabel *, "join-domain");
+  join_name = OBJ (GtkEntry *, "join-name");
+  join_password = OBJ (GtkEntry *, "join-password");
+  join_computer = OBJ (GtkEntry *, "join-computer");
+
+  gtk_entry_set_text (join_password, "");
+  gtk_widget_grab_focus (GTK_WIDGET (join_password));
+
+  kerberos = um_realm_object_get_kerberos (page->priv->realm);
+  membership = um_realm_object_get_kerberos_membership (page->priv->realm);
+
+  gtk_label_set_text (join_domain,
+                      um_realm_kerberos_get_domain_name (kerberos));
+
+  if (gethostname (hostname, sizeof (hostname)) == 0)
+    gtk_entry_set_text (join_computer, hostname);
+
+  clear_entry_validation_error (join_name);
+  clear_entry_validation_error (join_password);
+
+  if (!page->priv->join_prompted) {
+    name = um_realm_kerberos_membership_get_suggested_administrator (membership);
+    if (name && !g_str_equal (name, "")) {
+      g_debug ("Suggesting admin user: %s", name);
+      gtk_entry_set_text (join_name, name);
+    } else {
+      gtk_widget_grab_focus (GTK_WIDGET (join_name));
+    }
+
+  } else if (g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_HOSTNAME)) {
+    g_debug ("Bad host name: %s", error->message);
+    set_entry_validation_error (join_computer, error->message);
+
+  } else if (g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD)) {
+    g_debug ("Bad admin password: %s", error->message);
+    set_entry_validation_error (join_password, error->message);
+
+  } else {
+    g_debug ("Admin login failure: %s", error->message);
+    g_dbus_error_strip_remote_error (error);
+    set_entry_validation_error (join_name, error->message);
+  }
+
+  g_debug ("Showing admin password dialog");
+  gtk_window_set_transient_for (GTK_WINDOW (join_dialog),
+                                GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (page))));
+  gtk_window_set_modal (GTK_WINDOW (join_dialog), TRUE);
+  gtk_window_present (GTK_WINDOW (join_dialog));
+
+  page->priv->join_prompted = TRUE;
+  g_object_unref (kerberos);
+  g_object_unref (membership);
+
+  /* And now we wait for on_join_response() */
+}
+
+static void
+on_join_login (GObject *source,
+               GAsyncResult *result,
+               gpointer user_data)
+{
+  GisAccountPage *page = user_data;
+  GError *error = NULL;
+  GBytes *creds;
+
+  um_realm_login_finish (page->priv->realm, result, &creds, &error);
+
+  /* Logged in as admin successfully, use creds to join domain */
+  if (error == NULL) {
+    if (!um_realm_join_as_admin (page->priv->realm,
+                                 gtk_entry_get_text (OBJ (GtkEntry *, "join-name")),
+                                 gtk_entry_get_text (OBJ (GtkEntry *, "join-password")),
+                                 creds, NULL, on_realm_joined, page)) {
+      show_error_dialog (page, _("No supported way to authenticate with this domain"), NULL);
+      g_message ("Authenticating as admin is not supported by the realm");
+    }
+
+    g_bytes_unref (creds);
+
+  /* Couldn't login as admin, show prompt again */
+  } else {
+    join_show_prompt (page, error);
+    g_message ("Couldn't log in as admin to join domain: %s", error->message);
+    g_error_free (error);
+  }
+}
+
 static void
 on_realm_joined (GObject *source,
                  GAsyncResult *result,
                  gpointer user_data)
 {
   GisAccountPage *page = user_data;
-  UmRealmObject *realm = UM_REALM_OBJECT (source);
   GError *error = NULL;
 
-  um_realm_join_finish (realm, result, &error);
+  um_realm_join_finish (page->priv->realm, result, &error);
 
   /* Yay, joined the domain, register the user locally */
   if (error == NULL) {
     g_debug ("Joining realm completed successfully");
-    enterprise_permit_user_login (page, realm);
+    enterprise_permit_user_login (page, page->priv->realm);
 
     /* Credential failure while joining domain, prompt for admin creds */
   } else if (g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_LOGIN) ||
-             g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD)) {
-    g_debug ("Joining realm failed due to credentials");
+             g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD) ||
+             g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_HOSTNAME)) {
+    g_debug ("Joining realm failed due to credentials or host name");
 
-    /* XXX */
-    /* join_show_prompt (self, error); */
+    join_show_prompt (page, error);
 
     /* Other failure */
   } else {
     show_error_dialog (page, _("Failed to join domain"), error);
     g_message ("Failed to join the domain: %s", error->message);
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
   }
 
   g_clear_error (&error);
@@ -635,31 +835,40 @@ on_realm_login (GObject *source,
                 gpointer user_data)
 {
   GisAccountPage *page = user_data;
-  UmRealmObject *realm = UM_REALM_OBJECT (source);
   GError *error = NULL;
-  GBytes *creds;
+  GBytes *creds = NULL;
+
+  um_realm_login_finish (page->priv->realm, result, &creds, &error);
+
+  /*
+   * User login is valid, but cannot authenticate right now (eg: user needs
+   * to change password at next login etc.)
+   */
+  if (g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_CANNOT_AUTH)) {
+    g_clear_error (&error);
+    creds = NULL;
+  }
 
-  um_realm_login_finish (result, &creds, &error);
   if (error == NULL) {
 
     /* Already joined to the domain, just register this user */
-    if (um_realm_is_configured (realm)) {
+    if (um_realm_is_configured (page->priv->realm)) {
       g_debug ("Already joined to this realm");
-      enterprise_permit_user_login (page, realm);
+      enterprise_permit_user_login (page, page->priv->realm);
 
       /* Join the domain, try using the user's creds */
-    } else if (!um_realm_join_as_user (realm,
+    } else if (creds == NULL ||
+               !um_realm_join_as_user (page->priv->realm,
                                        gtk_entry_get_text (OBJ (GtkEntry *, "enterprise-login")),
                                        gtk_entry_get_text (OBJ (GtkEntry *, "enterprise-password")),
-                                       creds, NULL,
+                                       creds, page->priv->cancellable,
                                        on_realm_joined,
                                        page)) {
 
       /* If we can't do user auth, try to authenticate as admin */
       g_debug ("Cannot join with user credentials");
 
-      /* XXX: creds */
-      /* join_show_prompt (self, NULL); */
+      join_show_prompt (page, error);
     }
 
     g_bytes_unref (creds);
@@ -669,30 +878,33 @@ on_realm_login (GObject *source,
     g_debug ("Problem with the user's login: %s", error->message);
     set_entry_validation_error (OBJ (GtkEntry *, "enterprise-login"), error->message);
     gtk_widget_grab_focus (WID ("enterprise-login"));
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
 
   } else if (g_error_matches (error, UM_REALM_ERROR, UM_REALM_ERROR_BAD_PASSWORD)) {
     g_debug ("Problem with the user's password: %s", error->message);
     set_entry_validation_error (OBJ (GtkEntry *, "enterprise-password"), error->message);
     gtk_widget_grab_focus (WID ("enterprise-password"));
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
 
     /* Other login failure */
   } else {
     show_error_dialog (page, _("Failed to log into domain"), error);
     g_message ("Couldn't log in as user: %s", error->message);
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
   }
 
   g_clear_error (&error);
 }
 
 static void
-enterprise_check_login (GisAccountPage *page, UmRealmObject *realm)
+enterprise_check_login (GisAccountPage *page)
 {
-  g_assert (realm);
+  g_assert (page->priv->realm);
 
-  um_realm_login (realm,
+  um_realm_login (page->priv->realm,
                   gtk_entry_get_text (OBJ (GtkEntry *, "enterprise-login")),
                   gtk_entry_get_text (OBJ (GtkEntry *, "enterprise-password")),
-                  NULL,
+                  page->priv->cancellable,
                   on_realm_login,
                   page);
 }
@@ -714,8 +926,8 @@ on_realm_discover_input (GObject *source,
   if (error == NULL) {
     UmRealmObject *realm;
     g_assert (realms != NULL);
-    realm = g_object_ref (realms->data);
-    enterprise_check_login (page, realm);
+    priv->realm = g_object_ref (realms->data);
+    enterprise_check_login (page);
     g_list_free_full (realms, g_object_unref);
 
   } else {
@@ -724,6 +936,7 @@ on_realm_discover_input (GObject *source,
     g_message ("Couldn't discover domain: %s", error->message);
     gtk_widget_grab_focus (WID ("enterprise-domain-entry"));
     set_entry_validation_error (OBJ (GtkEntry*, "enterprise-domain-entry"), error->message);
+    gis_page_apply_complete (GIS_PAGE (page), FALSE);
     g_error_free (error);
   }
 }
@@ -736,19 +949,22 @@ enterprise_add_user (GisAccountPage *page)
   GtkTreeIter iter;
   GtkComboBox *domain = OBJ(GtkComboBox*, "enterprise-domain");
 
+  priv->join_prompted = FALSE;
+  g_clear_object (&priv->realm);
+
   /* Already know about this realm, try to login as user */
   if (gtk_combo_box_get_active_iter (domain, &iter)) {
     gtk_tree_model_get (gtk_combo_box_get_model (domain),
-                        &iter, 1, &realm, -1);
-    enterprise_check_login (page, realm);
+                        &iter, 1, &priv->realm, -1);
+    enterprise_check_login (page);
 
     /* Something the user typed, we need to discover realm */
   } else {
     um_realm_manager_discover (priv->realm_manager,
                                gtk_entry_get_text (OBJ (GtkEntry*, "enterprise-domain-entry")),
-                               NULL,
+                               priv->cancellable,
                                on_realm_discover_input,
-                               priv);
+                               page);
   }
 }
 
@@ -759,6 +975,7 @@ save_account_data (GisAccountPage *page)
   switch (priv->mode) {
   case UM_LOCAL:
     local_create_user (page);
+    gis_page_apply_complete (GIS_PAGE (page), TRUE);
     break;
   case UM_ENTERPRISE:
     enterprise_add_user (page);
@@ -768,31 +985,80 @@ save_account_data (GisAccountPage *page)
   }
 }
 
+static gchar *
+realm_get_name (UmRealmObject *realm)
+{
+  UmRealmCommon *common;
+  gchar *name;
+
+  common = um_realm_object_get_common (realm);
+  name = g_strdup (um_realm_common_get_name (common));
+  g_object_unref (common);
+
+  return name;
+}
+
+static gboolean
+model_contains_realm (GtkTreeModel *model,
+                      const gchar *realm_name)
+{
+  gboolean contains = FALSE;
+  GtkTreeIter iter;
+  gboolean match;
+  gchar *name;
+  gboolean ret;
+
+  ret = gtk_tree_model_get_iter_first (model, &iter);
+  while (ret) {
+    gtk_tree_model_get (model, &iter, 0, &name, -1);
+    match = (g_strcmp0 (name, realm_name) == 0);
+    g_free (name);
+    if (match) {
+      g_debug ("ignoring duplicate realm: %s", realm_name);
+      contains = TRUE;
+      break;
+    }
+    ret = gtk_tree_model_iter_next (model, &iter);
+  }
+
+  return contains;
+}
+
 static void
 enterprise_add_realm (GisAccountPage *page,
                       UmRealmObject  *realm)
 {
   GisAccountPagePrivate *priv = page->priv;
   GtkTreeIter iter;
-  UmRealmCommon *common;
   GtkComboBox *domain = OBJ(GtkComboBox*, "enterprise-domain");
-  GtkListStore *model = OBJ(GtkListStore*, "enterprise-realms-model");
+  GtkTreeModel *model = OBJ(GtkTreeModel*, "enterprise-realms-model");
+  gchar *name;
+  gboolean ret;
 
-  g_debug ("Adding new realm to drop down: %s",
-           g_dbus_object_get_object_path (G_DBUS_OBJECT (realm)));
+  name = realm_get_name (realm);
 
-  common = um_realm_object_get_common (realm);
+  /*
+   * Don't add a second realm if we already have one with this name.
+   * Sometimes realmd returns two realms for the same name, if it has
+   * different ways to use that realm. The first one that realmd
+   * returns is the one it prefers.
+   */
 
-  gtk_list_store_append (model, &iter);
-  gtk_list_store_set (model, &iter,
-                      0, um_realm_common_get_name (common),
-                      1, realm,
-                      -1);
+  if (!model_contains_realm (model, name)) {
+    gtk_list_store_append (GTK_LIST_STORE (model), &iter);
+    gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+                        0, name,
+                        1, realm,
+                        -1);
 
-  if (!priv->domain_chosen && um_realm_is_configured (realm))
-    gtk_combo_box_set_active_iter (domain, &iter);
+    if (!priv->domain_chosen && um_realm_is_configured (realm))
+      gtk_combo_box_set_active_iter (domain, &iter);
 
-  g_object_unref (common);
+    g_debug ("added realm to drop down: %s %s", name,
+             g_dbus_object_get_object_path (G_DBUS_OBJECT (realm)));
+  }
+
+  g_free (name);
 }
 
 static void
@@ -893,12 +1159,13 @@ toggle_mode (GtkToggleButton *button,
 }
 
 static void
-next_page_cb (GisAssistant *assistant,
-              GisPage      *which_page,
-              GisPage      *this_page)
+gis_account_page_apply (GisPage *gis_page,
+                        GCancellable *cancellable)
 {
-  if (which_page == this_page)
-    save_account_data (GIS_ACCOUNT_PAGE (this_page));
+  GisAccountPage *page = GIS_ACCOUNT_PAGE (gis_page);
+  g_clear_object (&page->priv->cancellable);
+  page->priv->cancellable = g_object_ref (cancellable);
+  save_account_data (page);
 }
 
 static void
@@ -940,6 +1207,8 @@ gis_account_page_constructed (GObject *object)
   g_signal_connect_after (confirm_entry, "focus-out-event",
                           G_CALLBACK (confirm_entry_focus_out), page);
 
+  g_signal_connect (WID("join-dialog"), "response",
+                    G_CALLBACK (on_join_response), page);
   g_signal_connect (WID("enterprise-domain"), "changed",
                     G_CALLBACK (on_domain_changed), page);
   g_signal_connect (WID("enterprise-login"), "changed",
@@ -947,8 +1216,6 @@ gis_account_page_constructed (GObject *object)
 
   priv->act_client = act_user_manager_get_default ();
 
-  g_signal_connect (assistant, "next-page", G_CALLBACK (next_page_cb), page);
-
   clear_account_page (page);
   update_account_page_status (page);
 
@@ -956,6 +1223,7 @@ gis_account_page_constructed (GObject *object)
 
   priv->action = gtk_toggle_button_new_with_mnemonic ("_Use Enterprise Login");
   g_signal_connect (priv->action, "toggled", G_CALLBACK (toggle_mode), page);
+  g_object_bind_property (page, "applying", priv->action, "sensitive", G_BINDING_INVERT_BOOLEAN);
   gtk_widget_show (priv->action);
   g_object_ref_sink (priv->action);
 
@@ -972,10 +1240,12 @@ gis_account_page_dispose (GObject *object)
   GisAccountPage *page = GIS_ACCOUNT_PAGE (object);
   GisAccountPagePrivate *priv = page->priv;
 
-  g_clear_object (&priv->act_user);
-  g_clear_object (&priv->act_client);
+  if (priv->realm_manager && priv->realmd_watch)
+    g_bus_unwatch_name (priv->realmd_watch);
   g_clear_object (&priv->realm_manager);
   g_clear_object (&priv->action);
+  g_clear_object (&priv->realm);
+  g_clear_object (&priv->cancellable);
 
   G_OBJECT_CLASS (gis_account_page_parent_class)->dispose (object);
 }
@@ -1001,6 +1271,7 @@ gis_account_page_class_init (GisAccountPageClass *klass)
   page_class->page_id = PAGE_ID;
   page_class->locale_changed = gis_account_page_locale_changed;
   page_class->get_action_widget = gis_account_page_get_action_widget;
+  page_class->apply = gis_account_page_apply;
   object_class->constructed = gis_account_page_constructed;
   object_class->dispose = gis_account_page_dispose;
 
