@@ -44,7 +44,7 @@
 #endif
 
 #include <act/act.h>
-
+#include <unistd.h>
 #include <egg-list-box.h>
 #include <libgd/gd.h>
 
@@ -73,6 +73,7 @@ struct _GisKeyboardPagePrivate {
         GtkWidget *show_layout;
         GtkWidget *input_scrolledwindow;
         guint n_input_rows;
+        guint gkbd_pid;
 
         GSettings *input_settings;
         GnomeXkbInfo *xkb_info;
@@ -111,12 +112,23 @@ gis_keyboard_page_finalize (GObject *object)
 
 static void localed_proxy_ready (GObject *source, GAsyncResult *res, gpointer data);
 static void setup_input_section (GisKeyboardPage *self);
+static void update_input (GisKeyboardPage *self);
+
+static void
+next_page_cb (GisAssistant *assistant,
+              GisPage      *which_page,
+              GisPage      *this_page)
+{
+        if (which_page == this_page)
+                update_input (GIS_KEYBOARD_PAGE (this_page));
+}
 
 static void
 gis_keyboard_page_constructed (GObject *object)
 {
         GisKeyboardPage *self = GIS_KEYBOARD_PAGE (object);
         GisKeyboardPagePrivate *priv = self->priv;
+        GisAssistant *assistant = gis_driver_get_assistant (GIS_PAGE (self)->driver);
 
         G_OBJECT_CLASS (gis_keyboard_page_parent_class)->constructed (object);
 
@@ -136,6 +148,8 @@ gis_keyboard_page_constructed (GObject *object)
                                   (GAsyncReadyCallback) localed_proxy_ready,
                                   self);
 
+        g_signal_connect (assistant, "next-page", G_CALLBACK (next_page_cb), self);
+
         gis_page_set_complete (GIS_PAGE (self), TRUE);
         gtk_widget_show (GTK_WIDGET (self));
 }
@@ -781,6 +795,7 @@ show_selected_layout (GisKeyboardPage *self)
         const gchar *layout;
         const gchar *variant;
         gchar *commandline;
+        gchar **argv = NULL;
 
         selected = egg_list_box_get_selected_child (EGG_LIST_BOX (priv->input_list));
         if (selected == NULL)
@@ -825,7 +840,28 @@ show_selected_layout (GisKeyboardPage *self)
                 commandline = g_strdup_printf ("gkbd-keyboard-display -l %s",
                                                layout);
 
-        g_spawn_command_line_async (commandline, NULL);
+        if (!g_shell_parse_argv (commandline,
+                            NULL,
+                            &argv,
+                            NULL))
+          goto out;
+
+        if (priv->gkbd_pid)
+          {
+            kill (priv->gkbd_pid, 9);
+            priv->gkbd_pid = 0;
+          }
+
+        g_spawn_async (NULL,
+                       argv,
+                       NULL,
+                       G_SPAWN_SEARCH_PATH,
+                       NULL,
+                       NULL,
+                       &priv->gkbd_pid,
+                       NULL);
+        g_strfreev (argv);
+  out:
         g_free (commandline);
 }
 
