@@ -91,7 +91,6 @@ padded_label_new (char *text)
         gtk_widget_set_margin_top (widget, 10);
         gtk_widget_set_margin_bottom (widget, 10);
         gtk_box_pack_start (GTK_BOX (widget), gtk_label_new (text), FALSE, FALSE, 0);
-        gtk_widget_show_all (widget);
         return widget;
 }
 
@@ -115,6 +114,7 @@ language_widget_new (const char *locale_id,
                      gboolean    is_extra)
 {
         gchar *locale_name, *locale_current_name, *locale_untranslated_name;
+        GtkWidget *checkmark;
         LanguageWidget *widget = g_new0 (LanguageWidget, 1);
 
         locale_name = gnome_get_language_from_locale (locale_id, locale_id);
@@ -128,9 +128,17 @@ language_widget_new (const char *locale_id,
         widget->locale_untranslated_name = locale_untranslated_name;
         widget->is_extra = is_extra;
 
+        /* We add a check on each side of the label to keep it centered. */
+        checkmark = gtk_image_new_from_icon_name ("object-select-symbolic", GTK_ICON_SIZE_MENU);
+        gtk_widget_show (checkmark);
+        gtk_widget_set_opacity (checkmark, 0.0);
+        gtk_box_pack_start (GTK_BOX (widget->box), checkmark, FALSE, FALSE, 0);
+        gtk_box_reorder_child (GTK_BOX (widget->box), checkmark, 0);
+
         widget->checkmark = gtk_image_new_from_icon_name ("object-select-symbolic", GTK_ICON_SIZE_MENU);
         gtk_box_pack_start (GTK_BOX (widget->box), widget->checkmark,
                             FALSE, FALSE, 0);
+        gtk_widget_show (widget->checkmark);
 
         g_object_set_data_full (G_OBJECT (widget->box), "language-widget", widget,
                                 language_widget_free);
@@ -151,7 +159,7 @@ sync_checkmark (GtkWidget *child,
 
         locale_id = user_data;
         should_be_visible = g_str_equal (widget->locale_id, locale_id);
-        gtk_widget_set_visible (widget->checkmark, should_be_visible);
+        gtk_widget_set_opacity (widget->checkmark, should_be_visible ? 1.0 : 0.0);
 }
 
 static void
@@ -167,9 +175,18 @@ static GtkWidget *
 more_widget_new (void)
 {
         GtkWidget *widget;
+        GtkWidget *arrow;
 
-        widget = padded_label_new ("…");
+        widget = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 10);
         gtk_widget_set_tooltip_text (widget, _("More…"));
+
+        arrow = gtk_image_new_from_icon_name ("view-more-symbolic", GTK_ICON_SIZE_MENU);
+        gtk_style_context_add_class (gtk_widget_get_style_context (arrow), "dim-label");
+        gtk_widget_set_margin_top (widget, 10);
+        gtk_widget_set_margin_bottom (widget, 10);
+        gtk_misc_set_alignment (GTK_MISC (arrow), 0.5, 0.5);
+        gtk_box_pack_start (GTK_BOX (widget), arrow, TRUE, TRUE, 0);
+
         return widget;
 }
 
@@ -210,7 +227,7 @@ add_languages (CcLanguageChooser  *chooser,
         gtk_container_add (GTK_CONTAINER (priv->language_list), priv->more_item);
         gtk_container_add (GTK_CONTAINER (priv->language_list), priv->no_results);
 
-        gtk_widget_show (priv->language_list);
+        gtk_widget_show_all (priv->language_list);
 }
 
 static void
